"""Basic functions to create cryptographic keys, and to encrypt string data.

Basic cryptographic functions to aid in key management for the user.
"""

import nacl.bindings
import logging
from datetime import datetime
from urllib.parse import urlparse

from didauth.headers import HeaderSigner
from didauth.base import SignerBase

from aries_cloudagent.wallet.crypto import (create_keypair, bytes_to_b58, validate_seed,
                                            verify_signed_message, b58_to_bytes, sign_message)

log = logging.getLogger(__name__)


def create_did(seed: bytes):
    """Creates a DID and verkey from the given seed.

    Args:
        seed (bytes): the secret seed to use in generating the verkey.

    Returns:
        tuple: `(did, verkey)`.
    """
    seed = validate_seed(seed)
    verkey, _ = create_keypair(seed)
    did = bytes_to_b58(verkey[:16])
    return did, bytes_to_b58(verkey)


def verify_signature(signature, verkey, message=None):
    """Verifies a signature cryptographically.

    Args:
        agent (str): name of the agent whose master key should be used.
        signature (str): hex-encoded signature that should be verified.
        verkey (str): public key to use in verifying the signature.
    """
    s, v = bytes.fromhex(signature), b58_to_bytes(verkey)
    if message is not None:
        if isinstance(message, bytes):
            return verify_signed_message(s + message, v)
        else:
            return verify_signed_message(s + message.encode("UTF-8"), v)
    else:
        return verify_signed_message(s, v)


def anon_crypt_message(message: bytes, pk: bytes) -> bytes:
    """Apply anon_crypt to a binary message.

    Args:
        message: The message to encrypt
        pk: The verkey to encrypt the message for

    Returns:
        The anon encrypted message
    """
    _pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(pk)
    enc_message = nacl.bindings.crypto_box_seal(message, _pk)
    return enc_message


def anon_decrypt_message(enc_message: bytes, pk: bytes, sk: bytes) -> bytes:
    """Apply anon_decrypt to a binary message.

    Args:
        enc_message: The encrypted message
        pk: public key that the message was encrypted for.
        sk: private key that the message was encrypted for.

    Returns:
        The decrypted message
    """
    _pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(pk)
    _sk = nacl.bindings.crypto_sign_ed25519_sk_to_curve25519(sk)

    message = nacl.bindings.crypto_box_seal_open(enc_message, _pk, _sk)
    return message


def generate_keys(seed):
    """Generates keys from a given seed.

    Args:
        seed (str): of `32` characters or less. 32 characters encodes a 128-bit seed. If fewer
            than `32` are specified, it will be front-padded with zeros.
    """
    if isinstance(seed, bytes):
        byte_seed = seed
    else:
        byte_seed = f"{seed:0<32}".encode()
    pk, sk = create_keypair(byte_seed)
    did, verkey = create_did(byte_seed)
    return {    
        "seed": bytes_to_b58(seed) if isinstance(seed, bytes) else seed,
        "did": did, 
        "public": verkey, 
        "secret": sk.hex()
    }


class ExplicitSigner(SignerBase):
    """Message signer that uses an explicit set of keys to sign messages.
    """
    algorithm = "ed25519"
    def __init__(self, keys):
        super(ExplicitSigner, self).__init__(self, "ed25519")
        self._pubkey = b58_to_bytes(keys["public"])
        self._secret = keys["secret"]
        self.vk = keys["public"]

    def _sign(self, data: bytes) -> bytes:
        log.debug(f"Signing message of length {len(data)} using {self.vk}.")
        result = sign_message(data, bytes.fromhex(self._secret))
        log.debug(f"Signing result is {result} from {data}")
        return result


def get_signed_headers(keys, url, method, headers=None):
    """Gets the didauth headers to sign for an HTTP request.
    """
    o = urlparse(url)
    _headers = headers or {}
    _headers.update({
        'Date': str(datetime.utcnow()),
        'Host': o.netloc,
        'User-Agent': f'ClearOSMobile/1.0',
    })
    header_list = ['(request-target)']
    header_list.extend(_headers.keys())
    url_path = o.path
    log.debug(f"Using {header_list} as target headers for signing at {url_path}.")
    log.debug(f"Using public keypair {keys['public']} to sign headers.")

    try:
        hdr_signer = HeaderSigner(keys["did"], ExplicitSigner(keys), header_list)
        signed_headers = hdr_signer.sign(_headers, method, url_path)
        assert 'authorization' in signed_headers
    except:
        log.exception("Signing headers for `didauth`.")
        return None
        
    return signed_headers


from clearlife.cli import get_default_args, main


__version__ = '0.1.0'  # also change in setup.py


class ClearLifeClient(object):
    """API Client for interacting with the `clearlifed` API
    endpoints. Note that this object is just a wrapper around
    the commands in :mod:`clearlifed.cli`.
    """
    def __init__(self, name) -> None:
        super().__init__()
        self.name = name
        self.args = get_default_args(name)

        self._otp = None
        self._app_keys = None


    @property
    def otp(self):
        """Registers this server application with `clearlifed` to
        obtain end-to-end encryption keys.
        """
        if self._otp is None:
            args = self.args.copy()
            args.update({
                "command": "register",
            })

            self._otp = main(args)

        return self._otp


    @property
    def app_keys(self):
        if self._app_keys is None:
            args = self.args.copy()
            args.update({
                "command": "appkeys",
                "seed": self.otp["seed"]
            })
            self._app_keys = main(args)

        return self._app_keys


    def derive(self, context, keyid):
        """Derives the keypair and seed for the given context
        and keyid.
        """
        args = self.args.copy()
        args.update({
            "command": "derive",
            "context": context,
            "keyid": keyid,
            "seed": self.otp["seed"]
        })

        return main(args)



from datetime import datetime
import logging
import nacl
from aries_cloudagent.wallet.util import b58_to_bytes, b64_to_bytes, bytes_to_b58
from urllib.request import parse_http_list
import multidict
from typing import Mapping

log = logging.getLogger(__name__)


class VerifierException(Exception):
    pass


class SignerException(VerifierException):
    pass


def decode_string(key, enc_format: str) -> bytes:
    if enc_format == 'base58':
        return b58_to_bytes(key)
    elif enc_format == 'base64':
        return b64_to_bytes(key)
    elif enc_format == 'hex':
        return bytes.fromhex(key)
    else:
        raise VerifierException('Key format not supported: {}'.format(enc_format))


class VerifierBase:
    def __init__(self, _key_type, _pubkey=None):
        self._pubkey = None

    @property
    def public_key(self) -> bytes:
        pass

    def _verify(self, _message: bytes, _signature: bytes, android: bool = False) -> bool:
        raise SystemError('Not implemented')

    def verify(self, message, signature, signature_format=None, android=False) -> bool:
        if signature_format:
            signature = decode_string(signature, signature_format)
        return self._verify(message, signature, android)


class Verifier(VerifierBase):
    algorithm = 'ed25519'

    def __init__(self, key_type, pubkey):
        super(Verifier, self).__init__(key_type)
        self._pubkey = pubkey

    def _verify(self, message: bytes, signature: bytes, android: bool = False) -> bool:
        try:
            vk = bytes_to_b58(self._pubkey)
            if android:
                # Make sure the message we generated is in the signature payload,
                # then just verify the payload.
                log.debug(f"Android signature verification of {message} and {signature} for {vk}.")
                assert signature[-len(message):] == message
                nacl.bindings.crypto_sign_open(signature, self._pubkey)
            else:
                log.debug(f"Standard signature verification of {message} and {signature} for {vk}.")
                nacl.bindings.crypto_sign_open(signature + message, self._pubkey)
            return True
        except:
            log.debug("Error validating signature", exc_info=True)
            return False


class SignatureHandlers:
    def __init__(self):
        self._classes = {}

    def add_key_type(self, key_type: str, signer, verifier):
        self._classes[key_type] = (signer, verifier)

    def supports(self, key_type: str):
        return key_type in self._classes

    def get_supported(self):
        return self._classes.keys()

    def create_signer(self, key_type, secret, secret_format=None):
        if secret and secret_format:
            secret = decode_string(secret, secret_format)
        return self._classes[key_type][0](key_type, secret)

    def create_verifier(self, key_type, pubkey, pubkey_format=None):
        if pubkey and pubkey_format:
            pubkey = decode_string(pubkey, pubkey_format)
        return self._classes[key_type][1](key_type, pubkey)


def parse_authorization_header(header):
    if isinstance(header, bytes):
        header = header.decode('ascii')  # HTTP headers cannot be Unicode.

    auth = header.split(' ', 1)
    if len(auth) > 2:
        raise VerifierException('Invalid authorization header. (eg. Method ' +
                                'key1=value1,key2="value, \"2\"")')

    # Split up any args into a dictionary.
    values = multidict.CIMultiDict()
    if len(auth) == 2:
        auth_value = auth[1]
        if auth_value:
            # This is tricky string magic.  Let urllib do it.
            fields = parse_http_list(auth_value)

            for item in fields:
                # Only include keypairs.
                if '=' in item:
                    # Split on the first '=' only.
                    key, value = item.split('=', 1)
                    if not key or not value:
                        continue

                    # Unquote values, if quoted.
                    if value[0] == '"':
                        value = value[1:-1]

                    values[key] = value

    # ("Signature", {"headers": "date", "algorithm": "hmac-sha256", ... })
    return (auth[0], values)



def signing_header(name, values):
    if isinstance(values, str):
        value = values
    else:
        value = ', '.join(v.strip() for v in values)
    return '{}: {}'.format(name, value)


def generate_message(required_headers, headers, method=None, path=None) -> bytes:
    headers = multidict.CIMultiDict(headers)

    if not required_headers:
        required_headers = ['date']

    signable_list = []
    for h in required_headers:
        h = h.lower()
        if h == '(request-target)':
            if not method or not path:
                raise SignerException('Method and path arguments required when ' +
                                      'using "(request-target)"')
            signable_list.append(signing_header(h, '{} {}'.format(method.lower(), path)))
        else:
            if h not in headers:
                raise SignerException('Missing required header "{}"'.format(h))
            signable_list.append(signing_header(h, headers.getall(h)))

    signable = '\n'.join(signable_list).encode('ascii')
    return signable


class Ed25519HeaderVerifier:
    """
    Verifies an HTTP signature from given headers.
    """
    def __init__(self, pubkey: bytes, required_headers=None):
        self.pubkey = pubkey
        self._handlers = ALL
        log.debug(f"Required headers set as {required_headers}.")
        if required_headers is None:
            required_headers = ["date"] # implementors should require (request-target) and date
        self._required_headers = [h.lower() for h in required_headers]


    def verify(self, headers: Mapping, method=None, path=None, android=False):
        """
        Parse Signature Authorization header and verify signature

        `headers` is a dict or multidict of headers
        `host` is a override for the 'host' header (defaults to value in
            headers).
        `method` is the HTTP method (required when using '(request-target)').
        `path` is the HTTP path (required when using '(request-target)').

        Args:
            android (bool): when True, treat the signature as coming from an android 
                app running lazy-sodium.
        """
        if not 'authorization' in headers:
            return False

        auth_type, auth_params = parse_authorization_header(headers['authorization'])
        if auth_type.lower() != 'signature':
            return False

        for param in ('algorithm', 'keyId', 'signature'):
            if param not in auth_params:
                raise VerifierException("Unsupported HTTP signature, missing '{}'".format(param))

        auth_headers = (auth_params.get('headers') or 'date').lower().strip().split()

        missing_reqd = set(self._required_headers) - set(auth_headers)
        if missing_reqd:
            error_headers = ', '.join(missing_reqd)
            raise VerifierException(
                'One or more required headers not provided: {}'.format(error_headers))

        key_id, algo = auth_params['keyId'], auth_params['algorithm']

        if not self._handlers.supports(algo):
            raise VerifierException("Unsupported HTTP signature algorithm '{}'".format(algo))

        if not self.pubkey:
            raise VerifierException("Cannot locate public key for '{}'".format(key_id))
        log.debug("Got %s public key for '%s': %s", algo, key_id, self.pubkey)

        handler = self._handlers.create_verifier(algo, self.pubkey)
        message = generate_message(auth_headers, headers, method, path)

        signature = auth_params['signature']
        raw_signature = decode_string(signature, 'base64')

        if handler.verify(message, raw_signature, android=android):
            return {
                'verified': True,
                'algorithm': algo,
                'headers': auth_headers,
                'keyId': key_id,
                'key': self.pubkey,
                'signature': signature
            }
        raise VerifierException("Signature could not be verified for keyId '{}'".format(key_id))


ALL = SignatureHandlers()
ALL.add_key_type('ed25519', None, Verifier)


def auth_did_signature(did, pubkey, headers, method, url_path):
    """Checks the DID signature in the headers to make sure that it is valid.

    Args:
        headers (dict): headers from the request object.
        method (str): one of the HTTP methods used in the request.
        url_path (str): full request path (excluding the host name and query
            string parameter).
    """
    verifier = Ed25519HeaderVerifier(b58_to_bytes(pubkey))
    log.debug(f"Verifying headers: {headers} with {method} at {url_path}.")
    is_android = "okhttp" in headers.get("User-Agent", "") or "okhttp" in headers.get("user-agent", "")
    try:
        verified = verifier.verify(headers, method, url_path, android=is_android)
    except VerifierException:
        log.exception(f"Verifying DID-based signature for {headers} and {method}>{url_path}.")
        return False, None

    if isinstance(verified, bool) or "keyId" not in verified:
        return False, None

    their_did = verified["keyId"]
    did_ok = False
    if verified['verified'] and their_did == did:
        # Check that the datetime is within a 15 seconds
        did_ok = didauth_time_okay(headers["date"], 15)

    return did_ok, their_did


def didauth_time_okay(their_date, max_seconds):
    """Checks whether the time encoded in a `didauth` message is close enough to the timestamp on the server.

    Args:
        their_date (str): date/time from the request.
        max_seconds (int): maximum number of seconds allowed for the time difference to be okay.

    Returns:
        bool: `True` if the time is within `max_seconds` of the server time.
    """
    server_now = datetime.utcnow()
    request_now = datetime.strptime(their_date, "%Y-%m-%d %H:%M:%S.%f")

    if server_now > request_now:
        delta = server_now - request_now
    else:
        delta = request_now - server_now

    result = delta.seconds < max_seconds
    if not result:
        log.debug(f"DID auth failed; too long: {delta.seconds} > {max_seconds}\n"
                    f"\t{server_now}\n"
                    f"\t{request_now}")
        
    return result



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
$ python setup.py register sdist upload

First Time register project on pypi
https://pypi.org/manage/projects/

More secure to use twine to upload
$ pip3 install twine
$ python3 setup.py sdist
$ twine upload dist/clearlifed-0.0.1.tar.gz

Best practices for setup.py and requirements.txt
https://caremad.io/posts/2013/07/setup-vs-requirement/
"""


from glob import glob
from os.path import basename
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


setup(
    name='clearlife',
    version='0.1.0',  #  also change in src/clearlifed/__init__.py
    license='AGPL3',
    description='ClearLIFE Server SDK',
    long_description="Key management using derived keys for server applications.",
    author='',
    author_email='',
    url='https://gitlab.com/clearos/clearfoundation/clearlife-sdk',
    packages=['clearlife'],
    py_modules=[splitext(basename(path))[0] for path in glob('clearlife/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        # uncomment if you test on these interpreters:
        #'Programming Language :: Python :: Implementation :: PyPy',
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Utilities',
    ],
    python_requires='>=3.8.6',
    install_requires=[
        'termcolor==1.1.0',
        'cython==0.29.23',
        'aries_cloudagent==0.6.0',
        'pysodium==0.7.9',
    ],
    setup_requires=[
    ],
    entry_points={
        'console_scripts': [
            'clearlife = clearlife.cli:main'
        ]
    },
)




from os import path, chdir as os_chdir, getcwd, environ
from contextlib import contextmanager
from importlib import import_module
import logging, sys
from hashlib import sha256
from subprocess import Popen, PIPE
from enum import Enum
import base64

from . import msg

log = logging.getLogger(__name__)

reporoot = None
"""str: full path to the repository root for the code.
"""

def get_reporoot():
    """Gets the full path to the repository root.
    """
    import clearlife
    global reporoot
    if reporoot is None:
        packpath = path.abspath(clearlife.__file__)
        reporoot = path.dirname(path.dirname(packpath))

    return reporoot


get_reporoot()


@contextmanager
def chdir(target):
    """Context manager for executing some code within a different
    directory after which the current working directory will be set
    back to what it was before.

    Args:
        target (str): path to the directory to change into.
    """
    current = getcwd()
    try:
        os_chdir(target)
        yield target
    finally:
        os_chdir(current)


def relpath(s, reldir=None):
    """Returns the *repository-relative* path for the string `s`.

    Args:
        s (str): repository-relative path, see the examples.
        reldir (str): when specified, make the path relative to this folder
          instead of the *code* repository root.

    Examples:
        Suppose I have a repository at `/usr/opt/repo`, then:

        >>> relpath("./tests") == "/usr/opt/repo/tests"
        True
        >>> relpath("../other/docs") == "/usr/opt/other/docs"
        True
    """
    if reldir is None:
        reldir = reporoot
    with chdir(reldir):
        result = path.abspath(s)
    return result


def import_fqdn(fqdn):
    """Returns the object from the specified FQDN. Any exceptions raised will
    bubble up.
    Args:
        fqdn (str): '.'-separated list of `package.module.callable` to
          import. The callable will *not* be called.
    Returns:
        tuple: `(module, callable)`, where `module` is the module object that
        `callable` resides in.
    """
    # from importlib import import_module
    log.debug(f"Importing {fqdn} dynamically.")
    parts = fqdn.split('.')
    call = parts[-1]
    module = '.'.join(parts[0:-1])
    try:
        module = import_module(module)
    except ImportError:
        module = import_module(parts[0])
        if len(parts) > 2:
            for part in parts[1:-1]:
                module = getattr(module, part)
            
    call = getattr(module, call)
    return (module, call)


def execute(args, folder=None, wait=True, nlines=100, venv=None,
            printerr=True, env_vars=None, errignore=None, shell=True, **kwargs):
    """Executes the specified tuple that should include the command as
    first item and additional arguments afterward. See the
    documentation for :class:`subprocess.Popen` for details.
    Args:
        args (list): of `str`; first item should be command to
          execute; additional arguments following.
        folder (str): directory to switch into before executing the
          command.
        wait (bool): when True, block the current thread until
          execution completes; otherwise, returns immediately.
        nlines (int): by default, `stdout` and `stderr` are redirected to
          :data:`subprocess.PIPE`. This is the maximum number of lines that will
          be returned for large outputs (so that memory doesn't get overwhelmed
          by large outputs).
        venv (str): when not `None`, the name of a virtualenv to
          activate before running the command.
        printerr (bool): when True, if `stderr` is not empty, print
          the lines automatically.
        env_vars (dict): of environment variables to set before calling the
          execution. The variables will be set back after execution.
        errignore (str): if errors are produced that include this pattern, then
          they will *not* be printed to `stdout`.
        kwargs (dict): additional arguments that are passed directly
          to the :class:`subprocess.Popen` constructor.
    Returns:
        dict: with keys ['process', 'stdout', 'stderr'], where 'process' is the
        instance of the subprocess that was created; 'stdout' and 'stderr' are
        only included if they were set to :data:`subprocess.PIPE`.
    .. note:: If the output from 'stdout' and 'stderr' are too large, only the
      first 100 lines will be returned. Use parameter `nlines` to control output
      size.
    """
    # from subprocess import Popen, PIPE
    if "stdout" not in kwargs:
        kwargs["stdout"] = PIPE
    if "stderr" not in kwargs:
        kwargs["stderr"] = PIPE
    if folder is not None:
        kwargs["cwd"] = folder

    if venv is not None: # pragma: no cover No guarantee that virtual
                         # envs exist on testing machine.
        prefix = path.dirname(sys.executable)
        if isinstance(venv, str):
            base_home = environ.get("WORKON_HOME")
            if base_home is not None:
                venv_home = path.abspath(path.expanduser(base_home))
                prefix = path.join(venv_home, venv, "bin")
        args[0] = path.join(prefix, args[0])

    oldvars = {}
    if env_vars is not None:
        for name, val in env_vars.items():
            oldvars[name] = environ[name] if name in environ else None
            environ[name] = val

    msg.std("Executing `{}` in {}.".format(' '.join(args), folder), 2)
    pexec = Popen(' '.join(args), shell=shell, executable="/bin/bash", **kwargs)

    if wait:
        sout, serr = pexec.communicate()
    else:
        sout, serr = b"", b""

    if env_vars is not None:
        #Set the environment variables back to what they used to be.
        for name, val in oldvars.items():
            if val is None:
                del environ[name]
            else:
                environ[name] = val

    #Redirect the output and errors so that we don't pollute stdout.
    output = None
    if kwargs["stdout"] is PIPE:
        output = []
        for line in sout.decode("ascii").split('\n'):
            #Filter non fatal exceptions such as future warnings. A full list can be found here
            # https://docs.python.org/3/library/exepctions.html#exception-hierarchy
            if line.strip() == "":
                continue

            if not ("FutureWarning" in line or "import" in line or "\x1b[0m" in line):
                output.append(line)
                if len(output) >= nlines:
                    break

    error = None
    if kwargs["stderr"] is PIPE and serr != '':
        error = []
        for line in serr.strip().decode("ascii").split('\n'):
            if (errignore is None or errignore not in line) and line.strip() != "":
                error.append(line)
            if len(error) >= nlines:
                break

        if printerr and len(error) > 0 and all([isinstance(i, str) for i in error]):
            msg.err(''.join(error))

    return {
        "process": pexec,
        "output": output,
        "error": error
    }


class AutoName(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


def keyhash(key):
    """Returns the bytes of the b85-encoded SHA256 hash for the key so that key
    sizes are standardized.

    Args:
        key (str): key string of arbitrary length.

    Returns:
        bytes: the SHA256 digest of `key`.
    """
    sha = sha256()
    sha.update(key.encode("utf8"))
    
    return base64.b85encode(sha.digest())